import logging

from queue import Queue

from typing import List, Optional, Any, Callable, Dict


from JuMonC import settings



logger = logging.getLogger(__name__)



class _taskSwitcher:
    pending_tasks: Queue = Queue(settings.PENDING_TASKS_SOFT_LIMIT + settings.MAX_THREADS_PER_TASK * settings.MAX_WORKER_THREADS + 1)
    
    def __init__(self) -> None:
        self.max_id = 0
        """Singelton class that switches taks depending on id."""
        self.switch_dic: Dict[int, Callable[..., None]] = {}
        
    def executeNextTask(self) -> None:
        data = self.pending_tasks.get()
        logging.debug("Executing task: %s", str(data))
        task = self.switch_dic.get(data[0], lambda data: logging.warning("Invalid taskID: %s", str(data[0])))
        task(data)
        self.pending_tasks.task_done()
        
    def addTask(self, data: Optional[List[Any]]) -> None:
        logging.debug("Adding task with data: %s", str(data))
        if data is not None:
            self.pending_tasks.put(data)
            return
        logging.info("added taks has no data")
    
    def addFunction(self, func: Callable[..., None]) -> int:
        logging.debug("Adding function: %s", str(func))
        self.max_id = self.max_id + 1
        self.switch_dic[self.max_id] = func
        logging.debug("function ID: %s", str(self.max_id))
        return self.max_id


task_switcher = _taskSwitcher()


import logging

from typing import Optional, Dict, Any, List
from types import ModuleType

from JuMonC.tasks import Plugin


logger = logging.getLogger(__name__)

class _NvidiaPlugin(Plugin.Plugin):
    
    _pynvml: Optional[ModuleType]
    _smi: Optional[ModuleType]
    
    _config = ["driver_version", "count", "name", "serial", "vbios_version",
              "display_mode", "persistence_mode", "compute_mode",
              "pci.bus_id", "pci.bus", "pcie.link.gen.max", "pcie.link.width.max",
              "clocks_throttle_reasons.supported", "clocks.max.graphics", "clocks.max.sm", "clocks.max.memory"
              "power.management", "power.limit", "power.min_limit", "power.max_limit",]
    _status = ["pcie.link.gen.current", "pcie.link.width.current",
              "fan.speed", "power.draw", "pstate",
              "memory.total", "memory.used", "memory.free",
              "utilization.gpu", "utilization.memory",
              "temperature.gpu", "temperature.memory",
              "clocks.current.graphic", "clocks.current.sm", "clocks.current.memory", "clocks.current.video",
              "clocks.applications.graphics", "clocks.applications.memory",
              "mig.mode.current", "mig.mode.pending"]
    
    def __init__(self) -> None:
        """Init pynvml!"""
        super().__init__()
        if self.isWorking():
            if isinstance(self._pynvml, ModuleType):
                try:
                    self._pynvml.nvmlInit()
                except self._pynvml.nvml.NVMLError_LibraryNotFound:
                    logging.warning("pynvml cannot find needed gpu libraries, therefore gpu functionality not avaiable")
                    self.works = False
                    return
                    
            if isinstance(self._smi, ModuleType):
                self.nvsmi = self._smi.smi.nvidia_smi.getInstance()
            else:
                logging.error("Something went wrong, in gpu plugin, pynvml is not ModuleType")

    
    def _isWorking(self) -> bool:
        try:
            self._pynvml = __import__('pynvml')
        except ModuleNotFoundError:
            logging.warning("pynvml can not be imported, therefore gpu functionality not avaiable")
            return False
        try:
            self._smi = __import__('pynvml.smi')
        except ModuleNotFoundError:
            logging.warning("pynvml.smi can not be imported, therefore gpu functionality not avaiable")
            return False
        return True
    
    def _getInfo(self, name:str) -> Dict[str, Any]:
        return self.nvsmi.DeviceQuery(name)
    
    def getConfigList(self) -> List[str]:
        return self._config
    
    def getStatusList(self) -> List[str]:
        return self._status
    
    def getConfigData(self, dataType:str) -> Dict[str, Any]:
        if dataType in self._config:
            return self._getInfo(dataType)
        return {}
    
    def getStatusData(self, dataType:str) -> Dict[str, Any]:
        if dataType in self._status:
            return self._getInfo(dataType)
        return {}
    
    
plugin = _NvidiaPlugin()


import logging

from mpi4py import MPI

from typing import Any, List, Optional, Tuple


from JuMonC.tasks import taskPool
from JuMonC.tasks.taskSwitcher import task_switcher
from JuMonC.models import dataStore



logger = logging.getLogger(__name__)



__comm = MPI.COMM_WORLD
__keep_running = True



def waitForCommands() -> None:
    taskPool.setupTaskPool()
    while __keep_running:
        task_switcher.executeNextTask()


def sendCommand(data: List[int]) -> None:
    __comm.bcast(data, root=0)


def gatherResult(dataID: int, command: List[int]) -> None:
    __comm.bcast(dataID, root=0)
    (rec_res_avai, result) = testResultAvaiable(dataID)
    
    if  rec_res_avai == 1:
        resultCom = __comm.reduce(result, op = MPI.SUM, root = 0)
        dataStore.removeResult(dataID)
        dataStore.addResult(dataID, resultCom)
    else: 
        logging.info("Gathering resheduled, command: %s", str(command))
        #taskSwitcher.tasks.add(command)

        
def testResultAvaiable(dataID: int) -> Tuple[int, Optional[Any]]:
    result_avaiable: int = 0
    result: Optional[Any] = None
    try:
        result = dataStore.getResult(dataID)
        result_avaiable = 1
    except KeyError:
        result_avaiable = 0

    rec_res_avai: int = 0
    __comm.Allreduce(result_avaiable, rec_res_avai, op=MPI.PROD)
    return (rec_res_avai, result)


"""
The handlers package of JuMonC provides the handlers to interact with JuMonC using the REST-API.

The REST API is documented with openapi, and rendered well by gitlab on:
https://gitlab.jsc.fz-juelich.de/witzler1/jumonc/-/blob/CI_fixes/doc/REST_API/openapi.yaml
"""

import logging

from flask import jsonify, make_response, Response

from typing import List, Any

import sys

from mpi4py import MPI

from JuMonC.handlers.base import RESTAPI
from JuMonC._version import __version__
from JuMonC.tasks.taskSwitcher import task_switcher
from JuMonC.authentication import scopes
from JuMonC.authentication.check import check_auth


logger = logging.getLogger(__name__)


@RESTAPI.route("/ping", methods=["GET"])
def ping() -> str:
    return "PONG.\n"



# stop all MPI handlers
def MPI_fin(data:List[Any]) -> None:
    logging.info("Stopping JuMonC")
    logging.debug("data: %s", str(data))
    MPI.Finalize()
    sys.exit(1)


stop_handlers_id = task_switcher.addFunction(MPI_fin)

@RESTAPI.route("/stop")
@check_auth(scopes["full"])
def stopJuMonC() -> Response:
    logging.info("Stopping JuMonC")
    task_switcher.addTask([stop_handlers_id])

    return make_response(jsonify("Stopping server and shutting down JuMonC\n"), 200)


@RESTAPI.route("/version", methods=["GET"])
def version() -> Response:
    response_body = {"JuMonC_version": __version__}
    return make_response(jsonify(response_body), 200)


@RESTAPI.route("/all_links", methods=["GET"])
@check_auth(scopes["full"])
def allLinks() -> Response:
    return make_response(jsonify(RESTAPI.url_map), 200)